# Copyright 2024 Agnostiq Inc.
"""A collection of reusable blueprint types for creating installed
blueprints."""

from .single_service_deployment import SingleServiceBlueprint
from .single_service_workflow import ServiceWorkflowBlueprint
